@@ -1,6 +1,7 @@
 /* GIO - GLib Input, Output and Streaming Library
  * 
- * Copyright (C) 2008 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2008,2009 Benjamin Otte <otte@gnome.org>
+ *               2008,2009 Andreas Henriksson <andreas@fatal.se>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,7 +18,8 @@
  * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  * Boston, MA 02111-1307, USA.
  *
- * Author: Benjmain Otte <otte@gnome.org>
+ * Author: Benjamin Otte <otte@gnome.org>
+ *         Andreas Henriksson <andreas@fatal.se>
  */
 
 
@@ -346,7 +348,6 @@ ftp_connection_receive (FtpConnection *conn,
 			ResponseFlags  flags)
 {
   SoupSocketIOStatus status;
-  gsize n_bytes;
   gboolean got_boundary;
   char *last_line;
   enum {
@@ -364,63 +365,62 @@ ftp_connection_receive (FtpConnection *conn,
   conn->read_bytes = 0;
   while (reply_state != DONE)
     {
-      if (conn->read_buffer_size - conn->read_bytes < 128)
-	{
-	  gsize new_size = conn->read_buffer_size + 1024;
-	  /* FIXME: upper limit for size? */
-	  gchar *new = g_try_realloc (conn->read_buffer, new_size);
-	  if (new)
-	    {
-	      conn->read_buffer = new;
-	      conn->read_buffer_size = new_size;
-	    }
-	  else
-	    {
-	      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-				   _("Invalid reply"));
-	      return 0;
-	    }
-	}
       last_line = conn->read_buffer + conn->read_bytes;
-      status = soup_socket_read_until (conn->commands,
-				       last_line,
-				       /* -1 byte for nul-termination */
-				       conn->read_buffer_size - conn->read_bytes - 1,
-				       "\r\n",
-				       2,
-				       &n_bytes,
-				       &got_boundary,
-				       conn->job->cancellable,
-				       &conn->error);
+      do {
+        gsize n_bytes;
 
-      conn->read_bytes += n_bytes;
-      conn->read_buffer[conn->read_bytes] = 0;
-      DEBUG ("<-- %s", last_line);
+        /* the available size must at least allow for boundary size (2)
+         * bytes to be available for reading */
+        if (conn->read_buffer_size - conn->read_bytes < 128)
+          {
+            gsize new_size = conn->read_buffer_size + 1024;
+            /* FIXME: upper limit for size? */
+            gchar *new = g_try_realloc (conn->read_buffer, new_size);
+            if (new)
+              {
+                /* make last line relative to new allocation */
+                last_line = new + (last_line - conn->read_buffer);
+                conn->read_buffer = new;
+                conn->read_buffer_size = new_size;
+              }
+            else
+              {
+                g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                                     _("Invalid reply"));
+                return 0;
+              }
+          }
+        status = soup_socket_read_until (conn->commands,
+                                         conn->read_buffer + conn->read_bytes,
+                                         /* -1 byte for nul-termination */
+                                         conn->read_buffer_size - conn->read_bytes - 1,
+                                         "\r\n",
+                                         2,
+                                         &n_bytes,
+                                         &got_boundary,
+                                         conn->job->cancellable,
+                                         &conn->error);
 
-      switch (status)
-	{
-	  case SOUP_SOCKET_OK:
-	  case SOUP_SOCKET_EOF:
-	    if (got_boundary)
-	      break;
-	    if (n_bytes > 0)
-	      continue;
+        conn->read_bytes += n_bytes;
+        conn->read_buffer[conn->read_bytes] = 0;
+
+        g_assert (status != SOUP_SOCKET_WOULD_BLOCK);
+        if (status == SOUP_SOCKET_ERROR)
+          return 0;
+
+        if (n_bytes == 0)
+          {
 	    g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
 				 _("Invalid reply"));
-	    /* fall through */
-	  case SOUP_SOCKET_ERROR:
-	    conn->read_buffer[conn->read_bytes] = 0;
-	    return 0;
-	  case SOUP_SOCKET_WOULD_BLOCK:
-	  default:
-	    g_assert_not_reached ();
-	    break;
-	}
+            return 0;
+          }
+      } while (!got_boundary);
+
+      DEBUG ("<-- %s", last_line);
 
       if (reply_state == FIRST_LINE)
 	{
-	  if (n_bytes < 4 ||
-	      last_line[0] <= '0' || last_line[0] > '5' ||
+	  if (last_line[0] <= '0' || last_line[0] > '5' ||
 	      last_line[1] < '0' || last_line[1] > '9' ||
 	      last_line[2] < '0' || last_line[2] > '9')
 	    {
@@ -444,8 +444,9 @@ ftp_connection_receive (FtpConnection *conn,
 	}
       else
 	{
-	  if (n_bytes >= 4 &&
-	      memcmp (conn->read_buffer, last_line, 3) == 0 &&
+	  if (last_line[0] == conn->read_buffer[0] &&
+              last_line[1] == conn->read_buffer[1] &&
+              last_line[2] == conn->read_buffer[2] &&
 	      last_line[3] == ' ')
 	    reply_state = DONE;
 	}
@@ -683,6 +684,8 @@ ftp_connection_parse_features (FtpConnection *conn)
 	    }
 	}
     }
+  
+  g_strfreev (supported);
 }
 
 /* NB: you must free the connection if it's in error returning from here */
@@ -737,7 +740,7 @@ ftp_connection_login (FtpConnection *conn,
        *   problems with our client.
        * - we don't want to drown in spam.
        */
-      if (password == NULL)
+      if (password == NULL || password[0] == 0)
 	password = "gvfsd-ftp-" VERSION "@example.com";
       status = ftp_connection_send (conn, 0,
 				    "PASS %s", password);
@@ -904,7 +907,6 @@ ftp_connection_ensure_data_connection_pasv (FtpConnection *conn)
 {
   guint ip1, ip2, ip3, ip4, port1, port2;
   const char *s;
-  gboolean connected;
   SoupAddress *addr;
   guint status;
 
